# [[file:../../org/pyretechnics.org::fire-spread-functions][fire-spread-functions]]
# TODO: Compare this with numpy.random (is there a generator object that I can use here?)
import random
import numpy as np


def supply_layer(layer_dict, layer_name, layer_shape, layer_type, initial_value, overwrite_layers):
    layer = layer_dict.get(layer_name)
    if layer:
        if overwrite_layers:
            return layer
        else:
            return layer.copy()
    else:
        return np.full(layer_shape, initial_value, dtype=layer_type)


# FIXME: stub
def burn_cells(world_state):
    """
    1. Validate inputs
       - Make sure that all required parameters are present in config_dict
       - Make sure that all required entries are present in layer_dict
       - Check for entries in output_layer_dict; if present, copy/use as new output layers
         - Use time of arrival raster to identify the burn scar(s); if missing, we are simulating point ignitions
       - Make sure that ignited_cells_set is a set of integer 3-tuples
    2. Iterate through all cells in (rows, cols), compute the maximum spread rate and severity values, and store them in output arrays
       - Run surface and crown equations per cell
    3. Return a dictionary of the outputs
    """
    simulation_clock  = world_state["simulation_clock"]
    config_dict       = world_state["config_dict"]
    input_layer_dict  = world_state["input_layer_dict"]
    output_layer_dict = world_state.get("output_layer_dict", {})
    ignited_cells_set = world_state["ignited_cells_set"]

    # The random seed affects input layer perturbations.
    if config_dict.get("random_seed"):
        random.seed(config_dict.get("random_seed"))

    return None


# FIXME: stub
def grow_fire(world_state):
    """
    1. Validate inputs
       - Make sure that all required parameters are present in config_dict
       - Make sure that all required entries are present in layer_dict
       - Check for entries in output_layer_dict; if present, copy/use as new output layers
         - Use time of arrival raster to identify the burn scar(s); if missing, we are simulating point ignitions
       - Make sure that ignited_cells_set is a set of integer 3-tuples
    2. Iterate through all cells in ignited_cells_set, compute the perimeter-oriented spread rate and severity values, and store them in output arrays
    3. Perform constant spread out over the landscape in all directions
       - Run surface, crown, and spot equations per ignited cell
    4. Record the time_of_arrival in each cell as it spreads
    5. Continue until a stop condition is met (e.g., max_burn_duration or max_burned_area)
    6. Return a dictionary of the outputs
    """
    simulation_clock  = world_state["simulation_clock"]
    config_dict       = world_state["config_dict"]
    input_layer_dict  = world_state["input_layer_dict"]
    output_layer_dict = world_state.get("output_layer_dict", {})
    ignited_cells_set = world_state["ignited_cells_set"]

    # The random seed affects input layer perturbations and spotting calculations.
    if config_dict.get("random_seed"):
        random.seed(config_dict.get("random_seed"))

    # GridFire initializes the following 2D arrays for its spread algorithm:
    #
    #   :burn-time-matrix                [float32] time_of_arrival (with -1 for values in the interior of the burn scar)
    #   :eccentricity-matrix             [float32] -1 in burn scar, 0 otherwise
    #   :fireline-intensity-matrix       [float32] -1 in burn scar, 0 otherwise
    #   :fire-spread-matrix              [float32] +1 in burn scar, 0 otherwise
    #   :fire-type-matrix                [float32] -1 in burn scar, 0 otherwise
    #   :firebrand-count-matrix          [ int32 ]  0 everywhere (when spotting params are passed)
    #   :flame-length-matrix             [float32] -1 in burn scar, 0 otherwise
    #   :directional-flame-length-matrix [float32] -1 in burn scar, 0 otherwise (when compute-directional-values? = true)
    #   :max-spread-direction-matrix     [float32] -1 in burn scar, 0 otherwise
    #   :max-spread-rate-matrix          [float32] -1 in burn scar, 0 otherwise
    #   :modified-time-matrix            [ int32 ]  0 everywhere
    #   :residence-time-matrix           [float32] -1 in burn scar, 0 otherwise (when compute-directional-values? = true)
    #   :reaction-intensity-matrix       [float32] -1 in burn scar, 0 otherwise (when compute-directional-values? = true)
    #   :spot-matrix                     [float32]  0 everywhere (when spotting params are passed)
    #   :spread-rate-matrix              [float32] -1 in burn scar, 0 otherwise
    #   :spread-rate-sum-matrix          [float32]  0 everywhere (when compute-directional-values? = true)
    #   :travel-lines-matrix             [ int16 ]  0 everywhere
    #   :x-magnitude-sum-matrix          [float32]  0 everywhere (when compute-directional-values? = true)
    #   :y-magnitude-sum-matrix          [float32]  0 everywhere (when compute-directional-values? = true)
    #
    # Equivalent Pyretechnics 2D arrays in output_layer_dict:
    #
    #   ========================= Output 2D Arrays =========================
    #   time_of_arrival                  :burn-time-matrix
    #   max_surface_spread_direction     :max-spread-direction-matrix
    #   max_crown_spread_direction       :max-spread-direction-matrix
    #   perimeter_spread_direction       N/A
    #   max_surface_spread_rate          :max-spread-rate-matrix
    #   max_crown_spread_rate            :max-spread-rate-matrix
    #   perimeter_spread_rate            :spread-rate-matrix
    #   max_surface_fireline_intensity   :fireline-intensity-matrix
    #   max_crown_fireline_intensity     :fireline-intensity-matrix
    #   perimeter_fireline_intensity     :fireline-intensity-matrix
    #   max_surface_flame_length         :flame-length-matrix
    #   max_crown_flame_length           :flame-length-matrix
    #   perimeter_flame_length           :directional-flame-length-matrix
    #   fire_type                        :fire-type-matrix
    #   firebrand_ignition               :spot-matrix
    #
    #   ======================== Internal 2D Arrays ========================
    #   surface_eccentricity             :eccentricity-matrix
    #   crown_eccentricity               :eccentricity-matrix
    #
    # GridFire 2D arrays that are not needed by Pyretechnics:
    #
    #   ======================== Replaced 2D Arrays ========================
    #   :fire-spread-matrix              time_of_arrival
    #   :residence-time-matrix           max_surface_fireline_intensity, max_crown_fireline_intensity, surface_eccentricity, crown_eccentricity
    #   :reaction-intensity-matrix       max_surface_fireline_intensity, max_crown_fireline_intensity, surface_eccentricity, crown_eccentricity
    #
    #   ======================== Internal 2D Arrays ========================
    #   :modified-time-matrix            \
    #   :travel-lines-matrix              |
    #   :spread-rate-sum-matrix           |-- for its 2D spread algorithm
    #   :x-magnitude-sum-matrix           |
    #   :y-magnitude-sum-matrix          /

    (num_timesteps, num_rows, num_cols) = config_dict["simulation_shape"]
    layer_shape = (num_rows, num_cols)
    overwrite_outputs = config_dict["overwrite_outputs"]

    output_layer_dict = {
        "eulerian_level_set_phi_field"  : supply_layer(output_layer_dict, "eulerian_level_set_phi_field"  , layer_shape, "float16", np.nan, overwrite_outputs),
        "time_of_arrival"               : supply_layer(output_layer_dict, "time_of_arrival"               , layer_shape, "float32", np.nan, overwrite_outputs),
        "max_surface_spread_direction"  : supply_layer(output_layer_dict, "max_surface_spread_direction"  , layer_shape, "float16", np.nan, overwrite_outputs),
        "max_crown_spread_direction"    : supply_layer(output_layer_dict, "max_crown_spread_direction"    , layer_shape, "float16", np.nan, overwrite_outputs),
        "perimeter_spread_direction"    : supply_layer(output_layer_dict, "perimeter_spread_direction"    , layer_shape, "float16", np.nan, overwrite_outputs),
        "max_surface_spread_rate"       : supply_layer(output_layer_dict, "max_surface_spread_rate"       , layer_shape, "float16", np.nan, overwrite_outputs),
        "max_crown_spread_rate"         : supply_layer(output_layer_dict, "max_crown_spread_rate"         , layer_shape, "float16", np.nan, overwrite_outputs),
        "perimeter_spread_rate"         : supply_layer(output_layer_dict, "perimeter_spread_rate"         , layer_shape, "float16", np.nan, overwrite_outputs),
        "surface_eccentricity"          : supply_layer(output_layer_dict, "surface_eccentricity"          , layer_shape, "float16", np.nan, overwrite_outputs),
        "crown_eccentricity"            : supply_layer(output_layer_dict, "crown_eccentricity"            , layer_shape, "float16", np.nan, overwrite_outputs),
        "max_surface_fireline_intensity": supply_layer(output_layer_dict, "max_surface_fireline_intensity", layer_shape, "float32", np.nan, overwrite_outputs),
        "max_crown_fireline_intensity"  : supply_layer(output_layer_dict, "max_crown_fireline_intensity"  , layer_shape, "float32", np.nan, overwrite_outputs),
        "perimeter_fireline_intensity"  : supply_layer(output_layer_dict, "perimeter_fireline_intensity"  , layer_shape, "float32", np.nan, overwrite_outputs),
        "max_surface_flame_length"      : supply_layer(output_layer_dict, "max_surface_flame_length"      , layer_shape, "float16", np.nan, overwrite_outputs),
        "max_crown_flame_length"        : supply_layer(output_layer_dict, "max_crown_flame_length"        , layer_shape, "float16", np.nan, overwrite_outputs),
        "perimeter_flame_length"        : supply_layer(output_layer_dict, "perimeter_flame_length"        , layer_shape, "float16", np.nan, overwrite_outputs),
        "fire_type"                     : supply_layer(output_layer_dict, "fire_type"                     , layer_shape, "uint8"  ,      0, overwrite_outputs),
        "firebrand_ignition"            : supply_layer(output_layer_dict, "firebrand_ignition"                , layer_shape, "bool8"  ,  False, overwrite_outputs),
    }

    # RESUME at [[file:~/code/sig-gis/gridfire/src/gridfire/fire_spread.clj::(defn- run-loop]]
    # TODO: Investigate ELMFIRE's inputs to determine if we are missing anything needed by its API.

    return None
# fire-spread-functions ends here



"""
Custom data structures for keeping track of cells in the narrow band.

The narrow band is defined as the set of _tracked_ cells:
- a cell is _tracked_ iff it lies within a square buffer around a _frontier cell_;
- a _frontier cell_ is a cell that has a neighbor of opposite phi sign,
i.e. both cells lie on opposite sides of the fire front.

`NarrowBandTracker` maintains the set of tracked cells by mapping each track cell to a count,
the count of the frontier cells related to it.
The intended use is that the surrounding algorithm will follow the additions and deletions
to the set of frontier cells, and correspondingly call the functions
`inc_square_around()` and `dec_square_around()`.
`TrackedCellsIterator` allows to iterate over tracked cells
(more efficiently than a regular Python iterator).
"""
import cython
import cython as cy
from sortedcontainers import SortedDict
if cython.compiled:
    from cython.cimports.pyretechnics.cy_types import pyidx, coord_yx
else:
    from pyretechnics.py_types import pyidx, coord_yx

#===================================================================
# Class CellsCountSegment
#===================================================================

@cy.cclass
class CellsCountSegment:
    """
    For internal use by the NarrowBandTracker.

    A CellsCountSegment maps each (y, x) for x from x0 (included) to x0+16 (excluded)
    to a positive integer.
    """
    y     : pyidx
    x0    : pyidx
    counts: cy.ushort[16]


@cy.cfunc
@cy.exceptval(check=False)
def segment_is_empty(segment: CellsCountSegment) -> cy.bint:
    counts: cy.ushort[16] = segment.counts
    k: pyidx
    for k in range(16):
        if counts[k] > 0:
            return False
    return True


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def segment_is_pos_at(segment: CellsCountSegment, k: pyidx) -> cy.bint:
    return segment.counts[k] > 0


# FIXME: Why isn't this a class __init__ constructor? (Val: ensuring private functions, simplifying pxd declarations)
@cy.cfunc
def make_CellsCountSegment(y: pyidx, x0: pyidx) -> CellsCountSegment:
    """
    Creates an empty CellsCountSegment (with all counts set to zero).
    """
    ret: CellsCountSegment = CellsCountSegment()
    ret.y  = y
    ret.x0 = x0
    return ret


@cy.cfunc
def __find_first_k(segment: CellsCountSegment) -> pyidx:
    k: pyidx
    for k in range(16):
        if segment_is_pos_at(segment, k):
            return k
    raise ValueError("Segment is empty!")

#===================================================================
# Class NarrowBandTracker
#===================================================================

# TODO: OPTIM Maybe we can make this faster by making ys_list a raw array.
@cy.cclass
class NarrowBandTracker:
    """
    A custom data structure for keeping track of cells in the narrow band efficiently.
    Logically, this class acts as a sorted dict from (y, x) -> n,
    in which n is the number of frontier cells whose buffer contains (y, x).

    Invariants:
    - for all (y, x) in the keys, ys_offset <= y < ys_offset + len(ys_list)
    - ys_list[i] logically represents the row for y := ys_offset + i
    - ys_list[i] is either None or a non-empty SortedDict s
    - ys_list[i] is None iff the row is empty
    - the SortedDict s := ys_list[i] maps each (x // 16) value to a non-empty CellsCountSegment segment,
      i.e. one of the counts in segment.counts will be > 0
    - _rows_count is the number of entries in ys_list containing non-empty SortedDicts
    """
    num_tracked_cells: cy.int
    y_high           : pyidx
    x_high           : pyidx
    ys_offset        : pyidx
    ys_list          : list
    _rows_count      : cy.int


# FIXME: Why isn't this a class __init__ constructor? (Val: ensuring private functions, simplifying pxd declarations)
@cy.cfunc
def new_NarrowBandTracker(y_high: pyidx, x_high: pyidx) -> NarrowBandTracker:
    ret: NarrowBandTracker = NarrowBandTracker()
    ret.num_tracked_cells  = 0
    ret.y_high             = y_high
    ret.x_high             = x_high
    return ret


@cy.cfunc
@cy.exceptval(check=False)
def __ensure_y(tracked_cells: NarrowBandTracker, y: pyidx) -> cy.void:
    """
    Ensures that index y is covered by ys_offset and ys_list.
    """
    # Ensure that the ys_offset and ys_list fields of tracked_cells are initialized
    if tracked_cells.ys_list is None:
        tracked_cells.ys_offset = y
        tracked_cells.ys_list   = [None] * 8
    # Grow the tracked_cells contents to the left or right if necessary
    n: cy.int = len(tracked_cells.ys_list)
    while y < tracked_cells.ys_offset:
        # Double the size, extending left
        tracked_cells.ys_list    = ([None] * n) + tracked_cells.ys_list
        tracked_cells.ys_offset -= n # FIXME: What happens when this becomes negative?
        n = 2*n
    while y >= tracked_cells.ys_offset + n:
        # Double the size, extending right
        tracked_cells.ys_list = tracked_cells.ys_list + ([None] * n)
        n = 2*n


# TODO: OPTIM list and dict lookup operations require us to allocate
#       tons of boxed integers on the heap. We can maybe make this
#       faster by caching those integers in a native array. That said,
#       I expect Python to keep doing reference-counting mutations
#       during function calls...
@cy.cfunc
@cy.exceptval(check=False)
def inc_y_segment(tracked_cells: NarrowBandTracker, y: pyidx, x_start: pyidx, segment_length: cy.int) -> cy.void:
    __ensure_y(tracked_cells, y)
    y_idx: int        = y - tracked_cells.ys_offset
    s    : SortedDict = tracked_cells.ys_list[y_idx]
    if s is None:
        s                             = SortedDict()
        tracked_cells.ys_list[y_idx]  = s
        tracked_cells._rows_count    += 1
    xk       : pyidx                  = (x_start // 16) - 1
    segment0 : CellsCountSegment|None = None
    segment  : CellsCountSegment
    n_added  : cy.int                 = 0
    old_count: cy.ushort
    k        : pyidx
    i        : pyidx
    for i in range(segment_length):
        if (x_start + i) // 16 != xk:
            xk       = (x_start + i) // 16
            segment0 = s.get(xk)
            if segment0 is None:
                segment0 = make_CellsCountSegment(y, xk*16)
                s[xk]    = segment0
        segment            = segment0
        k                  = (x_start + i) % 16
        old_count          = segment.counts[k]
        segment.counts[k]  = old_count + 1
        n_added           += (old_count == 0)
    tracked_cells.num_tracked_cells += n_added


@cy.cfunc
def dec_y_segment(tracked_cells: NarrowBandTracker, y: pyidx, x_start: pyidx, segment_length: cy.int) -> cy.void:
    y_idx: int        = y - tracked_cells.ys_offset
    s    : SortedDict = tracked_cells.ys_list[y_idx]
    if s is None:
        raise ValueError("Empty SortedDict!")
    xk      : pyidx                  = (x_start // 16)
    segment0: CellsCountSegment|None = s.get(xk)
    if segment0 is None:
        raise ValueError("No CellsCountSegment found!")
    segment  : CellsCountSegment
    n_removed: cy.int = 0
    old_count: cy.ushort
    k        : pyidx
    i        : pyidx
    for i in range(segment_length):
        k                  = (x_start + i) % 16
        segment            = segment0
        old_count          = segment.counts[k]
        segment.counts[k]  = old_count - 1
        n_removed         += (old_count == 1)
        if segment_is_empty(segment):
            del s[xk]
        if (x_start + i + 1) // 16 != xk:
            xk       = (x_start + i + 1) // 16
            segment0 = s.get(xk)
    if len(s) == 0:
        tracked_cells.ys_list[y_idx]  = None
        tracked_cells._rows_count    -= 1
    tracked_cells.num_tracked_cells  -= n_removed


@cy.cfunc
@cy.exceptval(check=False)
def resolve_truncated_x_segment(tracked_cells: NarrowBandTracker,
                                x            : pyidx,
                                buffer_width : pyidx) -> tuple[pyidx, pyidx]:
    x_start        : pyidx = max(0, x - buffer_width)
    x_end          : pyidx = min(tracked_cells.x_high - 1, x + buffer_width)
    width_truncated: pyidx = x_end - x_start
    return (x_start, width_truncated)


@cy.cfunc
@cy.exceptval(check=False)
def inc_square_around(tracked_cells: NarrowBandTracker, y: pyidx, x: pyidx, buffer_width: pyidx) -> cy.void:
    """
    Mutates tracked_cells by incrementing all cells within `buffer_width` of (y, x).

    For example, if buffer_width = 3, then 7*7=49 counts are updated.
    """
    (x_start, width_truncated) = resolve_truncated_x_segment(tracked_cells, x, buffer_width)
    width: pyidx               = 2 * buffer_width + 1
    i    : pyidx
    for i in range(width):
        y1: pyidx = y - buffer_width + i
        if (y1 >= 0) and (y1 < tracked_cells.y_high):
            inc_y_segment(tracked_cells, y1, x_start, width_truncated)


@cy.cfunc
@cy.exceptval(check=False)
def dec_square_around(tracked_cells: NarrowBandTracker, y: pyidx, x: pyidx, buffer_width: pyidx) -> cy.void:
    """
    Mutates tracked_cells by decrementing all cells within `buffer_width` of (y, x).

    If buffer_width = 3, then 7*7=49 counts are updated.

    The mutated cells MUST all have a positive count before this function is called.
    """
    (x_start, width_truncated) = resolve_truncated_x_segment(tracked_cells, x, buffer_width)
    width: pyidx               = 2 * buffer_width + 1
    i    : pyidx
    for i in range(width):
        y1: pyidx = y - buffer_width + i
        if (y1 >= 0) and (y1 < tracked_cells.y_high):
            dec_y_segment(tracked_cells, y1, x_start, width_truncated)


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def nonempty_tracked_cells(tracked_cells: NarrowBandTracker) -> cy.bint:
    return tracked_cells.num_tracked_cells > 0


# Iterating over cells
# NOTE: @cy.cfunc is not applicable here, due to yield.
def iterate_segments(tracked_cells: NarrowBandTracker):
    """
    Returns a Python iterator over the CellsCountSegment-s in `tracked_cells`.
    """
    ys_list: list = tracked_cells.ys_list
    if ys_list is not None:
        for s in ys_list:
            if s is not None:
                for segment in s.values():
                    yield segment


# TODO: OPTIM Maybe we could make this even faster by iterating
#       directly in NarrowBandTracker instead of relying on a Python
#       iterator, effectively reimplementing iterate_segments()
#       without yield. Basic testing suggests a 4x potential
#       efficiency gain by getting rid of yield!
@cy.cclass
class TrackedCellsIterator:
    """
    An "Iterator-pattern" object for iterating efficiently over the cells of a NarrowBandTracker.
    """
    segment_iter   : object # A Python Iterator of CellsCountSegment. INVARIANT None iff iteration is exhausted.
    current_segment: CellsCountSegment # INVARIANT never None
    current_k      : pyidx # INVARIANT points to the index for the next cell to return.


    # NOTE: This implementation relies on the fact that a CellsCountSegment is never empty.
    def __cinit__(self, segment_iter):
        self.segment_iter = segment_iter
        try:
            self.current_segment = next(segment_iter)
            self.current_k       = __find_first_k(self.current_segment)
        except StopIteration: # Only if the iterator started empty!
            self.segment_iter    = None
            self.current_segment = make_CellsCountSegment(-1, -1)
            self.current_k       = 0


    # TODO: OPTIM It may be faster to maintain an iteration index and compare it to num_tracked_cells.
    @cy.cfunc
    @cy.exceptval(check=False)
    def has_next(self) -> cy.bint:
        return self.segment_iter is not None


    # FIXME: Cython uninitialized warning
    @cy.cfunc
    @cy.exceptval(check=False)
    def next_cell(self) -> coord_yx:
        segment: CellsCountSegment = self.current_segment
        y      : pyidx             = segment.y
        x      : pyidx             = segment.x0 + self.current_k
        ret    : coord_yx          = (y, x)
        __resolve_next(self)
        return ret


@cy.cfunc
@cy.exceptval(check=False)
def __resolve_next(tc_iter: TrackedCellsIterator) -> cy.bint:
    # At the end of this, either segment_iter is None (exhausted),
    # or current_segment and current_k point to the next cell.
    # Returns True if iteration was exhausted and False otherwise.
    current_segment: CellsCountSegment = tc_iter.current_segment
    k              : pyidx             = tc_iter.current_k + 1
    while k < 16:
        if segment_is_pos_at(current_segment, k):
            tc_iter.current_k = k
            return False
        k += 1
    # We exhausted the segment, beginning the next one:
    try:
        tc_iter.current_segment = next(tc_iter.segment_iter)
    except StopIteration:
        tc_iter.segment_iter = None
        return True
    tc_iter.current_k = __find_first_k(tc_iter.current_segment)
    return False


@cy.cfunc
@cy.inline
def tracked_cells_iterator(tracked_cells: NarrowBandTracker) -> TrackedCellsIterator:
    return TrackedCellsIterator(iterate_segments(tracked_cells))



# [[file:../../org/pyretechnics.org::burn-cells-imports][burn-cells-imports]]
import cython
import cython as cy
import numpy as np
if cython.compiled:
    from cython.cimports.numpy import ndarray
    from cython.cimports.pyretechnics.cy_types import \
        pyidx, vec_xy, vec_xyz, coord_tyx, fclaarr, FuelModel, FireBehaviorMin, FireBehaviorMax, SpreadBehavior
    from cython.cimports.pyretechnics.space_time_cube import ISpaceTimeCube, to_positive_index_range
    import cython.cimports.pyretechnics.conversion as conv
    import cython.cimports.pyretechnics.vector_utils as vu
    import cython.cimports.pyretechnics.fuel_models as fm
    import cython.cimports.pyretechnics.surface_fire as sf
    import cython.cimports.pyretechnics.crown_fire as cf
else:
    from numpy import ndarray
    from pyretechnics.py_types import \
        pyidx, vec_xy, vec_xyz, coord_tyx, fclaarr, FuelModel, FireBehaviorMin, FireBehaviorMax, SpreadBehavior
    from pyretechnics.space_time_cube import ISpaceTimeCube, to_positive_index_range
    import pyretechnics.conversion as conv
    import pyretechnics.vector_utils as vu
    import pyretechnics.fuel_models as fm
    import pyretechnics.surface_fire as sf
    import pyretechnics.crown_fire as cf
# burn-cells-imports ends here
# [[file:../../org/pyretechnics.org::burn-cells-as-head-fire][burn-cells-as-head-fire]]
@cy.cfunc
@cy.inline
def SpreadBehavior_to_dict(sb: SpreadBehavior) -> dict:
    return {
        "fire_type"         : sb.fire_type,
        "spread_rate"       : sb.spread_rate,
        "spread_direction"  : sb.spread_direction,
        "fireline_intensity": sb.fireline_intensity,
        "flame_length"      : sb.flame_length,
    }


@cy.ccall
def burn_cell_as_head_fire(space_time_cubes      : dict[str, ISpaceTimeCube],
                           space_time_coordinate : coord_tyx,
                           use_wind_limit        : cy.bint = True,
                           surface_lw_ratio_model: str = "behave",
                           crown_max_lw_ratio    : cy.float = 1e10) -> dict:
    """
    Given these inputs:
    - space_time_cubes             :: dictionary of (Lazy)SpaceTimeCube objects with these cell types
      - slope                         :: rise/run
      - aspect                        :: degrees clockwise from North
      - fuel_model                    :: integer index in fm.fuel_model_table
      - canopy_cover                  :: 0-1
      - canopy_height                 :: m
      - canopy_base_height            :: m
      - canopy_bulk_density           :: kg/m^3
      - wind_speed_10m                :: km/hr
      - upwind_direction              :: degrees clockwise from North
      - fuel_moisture_dead_1hr        :: kg moisture/kg ovendry weight
      - fuel_moisture_dead_10hr       :: kg moisture/kg ovendry weight
      - fuel_moisture_dead_100hr      :: kg moisture/kg ovendry weight
      - fuel_moisture_live_herbaceous :: kg moisture/kg ovendry weight
      - fuel_moisture_live_woody      :: kg moisture/kg ovendry weight
      - foliar_moisture               :: kg moisture/kg ovendry weight
      - fuel_spread_adjustment        :: float >= 0.0 (Optional: defaults to 1.0)
      - weather_spread_adjustment     :: float >= 0.0 (Optional: defaults to 1.0)
    - space_time_coordinate        :: (t,y,x)
    - use_wind_limit               :: boolean (Optional)
    - surface_lw_ratio_model       :: "rothermel" or "behave" (Optional)
    - crown_max_lw_ratio           :: float > 0.0 (Optional)

    return a dictionary with these fire behavior values for the space-time coordinate (t,y,x):
    - fire_type          :: 0 (unburned), 1 (surface), 2 (passive_crown), or 3 (active_crown)
    - spread_rate        :: m/min
    - spread_direction   :: (x, y, z) unit vector on the slope-tangential plane
    - fireline_intensity :: kW/m
    - flame_length       :: m
    """
    #================================================================================================
    # Destructure the space_time_coordinate
    #================================================================================================

    (t, y, x) = space_time_coordinate

    #================================================================================================
    # Unpack the space_time_cubes dictionary
    #================================================================================================

    # Topography, Fuel Model, and Vegetation
    slope              : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["slope"]).get(t, y, x)
    aspect             : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["aspect"]).get(t, y, x)
    fuel_model_number  : cy.int   = cy.cast(cy.int, cy.cast(ISpaceTimeCube, space_time_cubes["fuel_model"]).get(t, y, x))
    canopy_cover       : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["canopy_cover"]).get(t, y, x)
    canopy_height      : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["canopy_height"]).get(t, y, x)
    canopy_base_height : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["canopy_base_height"]).get(t, y, x)
    canopy_bulk_density: cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["canopy_bulk_density"]).get(t, y, x)

    # Wind, Surface Moisture, and Foliar Moisture
    wind_speed_10m               : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["wind_speed_10m"]).get(t, y, x)
    upwind_direction             : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["upwind_direction"]).get(t, y, x)
    fuel_moisture_dead_1hr       : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_dead_1hr"]).get(t, y, x)
    fuel_moisture_dead_10hr      : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_dead_10hr"]).get(t, y, x)
    fuel_moisture_dead_100hr     : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_dead_100hr"]).get(t, y, x)
    fuel_moisture_live_herbaceous: cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_live_herbaceous"]).get(t, y, x)
    fuel_moisture_live_woody     : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_live_woody"]).get(t, y, x)
    foliar_moisture              : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["foliar_moisture"]).get(t, y, x)

    # Spread Rate Adjustments (Optional)
    fuel_spread_adjustment   : cy.float = (cy.cast(ISpaceTimeCube, space_time_cubes["fuel_spread_adjustment"]).get(t, y, x)
                                           if "fuel_spread_adjustment" in space_time_cubes
                                           else 1.0)                                         # float >= 0.0
    weather_spread_adjustment: cy.float = (cy.cast(ISpaceTimeCube, space_time_cubes["weather_spread_adjustment"]).get(t, y, x)
                                           if "weather_spread_adjustment" in space_time_cubes
                                           else 1.0)                                         # float >= 0.0
    spread_rate_adjustment   : cy.float = fuel_spread_adjustment * weather_spread_adjustment # float >= 0.0

    #================================================================================================
    # Check whether cell is burnable
    #================================================================================================

    fuel_model      : FuelModel
    maybe_fuel_model: FuelModel|None = fm.fuel_model_table.get(fuel_model_number)

    if maybe_fuel_model:
        fuel_model = maybe_fuel_model

    if (maybe_fuel_model is None or not(fuel_model.burnable)):
        # Encountered unknown or non-burnable fuel model

        #================================================================================================
        # Create a 3D unit vector pointing upslope on the slope-tangential plane
        #================================================================================================

        upslope_direction: cy.float = conv.opposite_direction(aspect)
        slope_vector_2d  : vec_xy   = conv.azimuthal_to_cartesian(slope, upslope_direction)
        slope_vector_3d  : vec_xyz  = vu.to_slope_plane(slope_vector_2d, slope_vector_2d)
        default_direction: vec_xyz  = (0.0, 1.0, 0.0) # default: North
        spread_direction : vec_xyz  = vu.as_unit_vector_3d(slope_vector_3d) if slope > 0.0 else default_direction

        #============================================================================================
        # Return zero surface fire behavior
        #============================================================================================

        return {
            "fire_type"         : 0, # unburned
            "spread_rate"       : 0.0,
            "spread_direction"  : spread_direction,
            "fireline_intensity": 0.0,
            "flame_length"      : 0.0,
        }

    else:
        # Encountered burnable fuel model

        #============================================================================================
        # Compute derived parameters
        #============================================================================================

        fuel_moisture               : fclaarr  = (fuel_moisture_dead_1hr,
                                                  fuel_moisture_dead_10hr,
                                                  fuel_moisture_dead_100hr,
                                                  0.0, # fuel_moisture_dead_herbaceous
                                                  fuel_moisture_live_herbaceous,
                                                  fuel_moisture_live_woody) # kg moisture/kg ovendry weight
        fuel_bed_depth              : cy.float = fuel_model.delta                      # ft
        heat_of_combustion          : cy.float = conv.Btu_lb_to_kJ_kg(fuel_model.h[0]) # kJ/kg
        estimated_fine_fuel_moisture: cy.float = fuel_moisture_dead_1hr                # kg moisture/kg ovendry weight

        #============================================================================================
        # Calculate midflame wind speed
        #============================================================================================

        # Convert from 10m wind speed to 20ft wind speed
        wind_speed_20ft: cy.float = conv.wind_speed_10m_to_wind_speed_20ft(wind_speed_10m) # km/hr

        # Convert 20ft wind speed from km/hr to m/min
        wind_speed_20ft_m_min: cy.float = conv.km_hr_to_m_min(wind_speed_20ft) # m/min

        # Convert from 20ft wind speed to midflame wind speed in m/min
        midflame_wind_speed: cy.float = sf.calc_midflame_wind_speed(wind_speed_20ft_m_min,       # m/min
                                                                    fuel_bed_depth,              # ft
                                                                    conv.m_to_ft(canopy_height), # ft
                                                                    canopy_cover)                # 0-1

        #============================================================================================
        # Calculate surface fire behavior in the direction of maximum spread
        #============================================================================================

        # Apply fuel moisture to fuel model
        moisturized_fuel_model: FuelModel = fm.moisturize(fuel_model, fuel_moisture)

        # TODO: Memoize calc_surface_fire_behavior_no_wind_no_slope
        # Calculate no-wind-no-slope surface fire behavior
        surface_fire_min: FireBehaviorMin = sf.calc_surface_fire_behavior_no_wind_no_slope(moisturized_fuel_model,
                                                                                           spread_rate_adjustment)

        # Calculate surface fire behavior in the direction of maximum spread
        surface_fire_max: FireBehaviorMax = sf.calc_surface_fire_behavior_max(surface_fire_min,
                                                                              midflame_wind_speed,
                                                                              upwind_direction,
                                                                              slope,
                                                                              aspect,
                                                                              use_wind_limit,
                                                                              surface_lw_ratio_model)

        # Simplify the surface fire behavior fields for future combination with the crown fire behavior values
        spread_direction       : vec_xyz        = surface_fire_max.max_spread_direction
        surface_fire_max_simple: SpreadBehavior = sf.calc_surface_fire_behavior_in_direction(surface_fire_max,
                                                                                             spread_direction)

        #============================================================================================
        # Determine whether the surface fire transitions to a crown fire
        #============================================================================================

        if cf.van_wagner_crown_fire_initiation(surface_fire_max_simple.fireline_intensity,
                                               canopy_cover,
                                               canopy_base_height,
                                               foliar_moisture):

            #========================================================================================
            # Calculate crown fire behavior in the direction of maximum spread
            #========================================================================================

            # Calculate crown fire behavior in the direction of maximum spread
            crown_fire_max: FireBehaviorMax = cf.calc_crown_fire_behavior_max(canopy_height, canopy_base_height,
                                                                              canopy_bulk_density, heat_of_combustion,
                                                                              estimated_fine_fuel_moisture,
                                                                              wind_speed_10m, upwind_direction,
                                                                              slope, aspect, crown_max_lw_ratio)

            # Simplify the crown fire behavior fields for future combination with the surface fire behavior values
            spread_direction     : vec_xyz        = crown_fire_max.max_spread_direction
            crown_fire_max_simple: SpreadBehavior = cf.calc_crown_fire_behavior_in_direction(crown_fire_max,
                                                                                             spread_direction)

            #========================================================================================
            # Calculate combined fire behavior in the direction of maximum spread
            #========================================================================================

            combined_fire_max: SpreadBehavior = cf.calc_combined_fire_behavior(surface_fire_max_simple,
                                                                               crown_fire_max_simple)

            #========================================================================================
            # Return the combined fire behavior in the direction of maximum spread
            #========================================================================================

            return SpreadBehavior_to_dict(combined_fire_max)

        else:

            #========================================================================================
            # Return the surface fire behavior in the direction of maximum spread
            #========================================================================================

            return SpreadBehavior_to_dict(surface_fire_max_simple)


# TODO: Make a more efficient version that avoids space_time_cubes dictionary lookups for each cell
@cy.ccall
def burn_all_cells_as_head_fire(space_time_cubes      : dict[str, ISpaceTimeCube],
                                t                     : pyidx,
                                y_range               : tuple[pyidx, pyidx]|None = None,
                                x_range               : tuple[pyidx, pyidx]|None = None,
                                use_wind_limit        : cy.bint = True,
                                surface_lw_ratio_model: str = "behave",
                                crown_max_lw_ratio    : cy.float = 1e10) -> dict:
    """
    Given these inputs:
    - space_time_cubes             :: dictionary of (Lazy)SpaceTimeCube objects with these cell types
      - slope                         :: rise/run
      - aspect                        :: degrees clockwise from North
      - fuel_model                    :: integer index in fm.fuel_model_table
      - canopy_cover                  :: 0-1
      - canopy_height                 :: m
      - canopy_base_height            :: m
      - canopy_bulk_density           :: kg/m^3
      - wind_speed_10m                :: km/hr
      - upwind_direction              :: degrees clockwise from North
      - fuel_moisture_dead_1hr        :: kg moisture/kg ovendry weight
      - fuel_moisture_dead_10hr       :: kg moisture/kg ovendry weight
      - fuel_moisture_dead_100hr      :: kg moisture/kg ovendry weight
      - fuel_moisture_live_herbaceous :: kg moisture/kg ovendry weight
      - fuel_moisture_live_woody      :: kg moisture/kg ovendry weight
      - foliar_moisture               :: kg moisture/kg ovendry weight
      - fuel_spread_adjustment        :: float >= 0.0 (Optional: defaults to 1.0)
      - weather_spread_adjustment     :: float >= 0.0 (Optional: defaults to 1.0)
    - t                            :: temporal integer index into the ISpaceTimeCube objects
    - y_range                      :: (min_y, max_y) spatial integer index into the ISpaceTimeCube objects (Optional)
    - x_range                      :: (min_x, max_x) spatial integer index into the ISpaceTimeCube objects (Optional)
    - use_wind_limit               :: boolean (Optional)
    - surface_lw_ratio_model       :: "rothermel" or "behave" (Optional)
    - crown_max_lw_ratio           :: float > 0.0 (Optional)

    return a dictionary with these keys:
    - fire_type          :: 2D byte array (0=unburned, 1=surface, 2=passive_crown, 3=active_crown)
    - spread_rate        :: 2D float array (m/min)
    - spread_direction   :: 2D float array (degrees clockwise from North)
    - fireline_intensity :: 2D float array (kW/m)
    - flame_length       :: 2D float array (m)
    """
    slope_cube  : ISpaceTimeCube      = cy.cast(ISpaceTimeCube, space_time_cubes["slope"])
    bands       : pyidx               = slope_cube.shape[0]
    rows        : pyidx               = slope_cube.shape[1]
    cols        : pyidx               = slope_cube.shape[2]
    grid_shape  : tuple[int, int]     = (rows, cols)
    y_range_real: tuple[pyidx, pyidx] = to_positive_index_range(y_range, rows)
    x_range_real: tuple[pyidx, pyidx] = to_positive_index_range(x_range, cols)
    min_y       : pyidx               = y_range_real[0]
    max_y       : pyidx               = y_range_real[1]
    min_x       : pyidx               = x_range_real[0]
    max_x       : pyidx               = x_range_real[1]

    if not(0 <= t < bands):
        raise ValueError("The t value is out of range of the space_time_cubes.")

    if not(0 <= min_y < max_y <= rows):
        raise ValueError("The y_range values are out of range of the space_time_cubes.")

    if not(0 <= min_x < max_x <= rows):
        raise ValueError("The x_range values are out of range of the space_time_cubes.")

    fire_type_matrix          : ndarray         = np.zeros(grid_shape, dtype="uint8")
    spread_rate_matrix        : ndarray         = np.zeros(grid_shape, dtype="float32")
    spread_direction_matrix   : ndarray         = np.zeros(grid_shape, dtype="float32")
    fireline_intensity_matrix : ndarray         = np.zeros(grid_shape, dtype="float32")
    flame_length_matrix       : ndarray         = np.zeros(grid_shape, dtype="float32")
    fire_type_memview         : cy.uchar[:,::1] = fire_type_matrix
    spread_rate_memview       : cy.float[:,::1] = spread_rate_matrix
    spread_direction_memview  : cy.float[:,::1] = spread_direction_matrix
    fireline_intensity_memview: cy.float[:,::1] = fireline_intensity_matrix
    flame_length_memview      : cy.float[:,::1] = flame_length_matrix

    y                    : pyidx
    x                    : pyidx
    space_time_coordinate: coord_tyx
    for y in range(min_y, max_y):
        for x in range(min_x, max_x):
            space_time_coordinate          = (t, y, x)
            spread_behavior                = burn_cell_as_head_fire(space_time_cubes,
                                                                    space_time_coordinate,
                                                                    use_wind_limit,
                                                                    surface_lw_ratio_model,
                                                                    crown_max_lw_ratio)
            fire_type_memview[y,x]          = spread_behavior["fire_type"]
            spread_rate_memview[y,x]        = spread_behavior["spread_rate"]
            spread_direction_memview[y,x]   = vu.spread_direction_vector_to_angle(spread_behavior["spread_direction"])
            fireline_intensity_memview[y,x] = spread_behavior["fireline_intensity"]
            flame_length_memview[y,x]       = spread_behavior["flame_length"]

    return {
        "fire_type"         : fire_type_matrix,
        "spread_rate"       : spread_rate_matrix,
        "spread_direction"  : spread_direction_matrix,
        "fireline_intensity": fireline_intensity_matrix,
        "flame_length"      : flame_length_matrix,
    }
# burn-cells-as-head-fire ends here
# [[file:../../org/pyretechnics.org::burn-cells-toward-azimuth][burn-cells-toward-azimuth]]
@cy.ccall
def burn_cell_toward_azimuth(space_time_cubes      : dict[str, ISpaceTimeCube],
                             space_time_coordinate : coord_tyx,
                             azimuth               : cy.float,
                             use_wind_limit        : cy.bint = True,
                             surface_lw_ratio_model: str = "behave",
                             crown_max_lw_ratio    : cy.float = 1e10) -> dict:
    """
    Given these inputs:
    - space_time_cubes             :: dictionary of (Lazy)SpaceTimeCube objects with these cell types
      - slope                         :: rise/run
      - aspect                        :: degrees clockwise from North
      - fuel_model                    :: integer index in fm.fuel_model_table
      - canopy_cover                  :: 0-1
      - canopy_height                 :: m
      - canopy_base_height            :: m
      - canopy_bulk_density           :: kg/m^3
      - wind_speed_10m                :: km/hr
      - upwind_direction              :: degrees clockwise from North
      - fuel_moisture_dead_1hr        :: kg moisture/kg ovendry weight
      - fuel_moisture_dead_10hr       :: kg moisture/kg ovendry weight
      - fuel_moisture_dead_100hr      :: kg moisture/kg ovendry weight
      - fuel_moisture_live_herbaceous :: kg moisture/kg ovendry weight
      - fuel_moisture_live_woody      :: kg moisture/kg ovendry weight
      - foliar_moisture               :: kg moisture/kg ovendry weight
      - fuel_spread_adjustment        :: float >= 0.0 (Optional: defaults to 1.0)
      - weather_spread_adjustment     :: float >= 0.0 (Optional: defaults to 1.0)
    - space_time_coordinate        :: (t,y,x)
    - azimuth                      :: degrees clockwise from North on the horizontal plane
    - use_wind_limit               :: boolean (Optional)
    - surface_lw_ratio_model       :: "rothermel" or "behave" (Optional)
    - crown_max_lw_ratio           :: float > 0.0 (Optional)

    return a dictionary with these fire behavior values for the space-time coordinate (t,y,x):
    - fire_type          :: 0 (unburned), 1 (surface), 2 (passive_crown), or 3 (active_crown)
    - spread_rate        :: m/min
    - spread_direction   :: (x, y, z) unit vector on the slope-tangential plane
    - fireline_intensity :: kW/m
    - flame_length       :: m
    """
    #================================================================================================
    # Destructure the space_time_coordinate
    #================================================================================================

    (t, y, x) = space_time_coordinate

    #================================================================================================
    # Unpack the space_time_cubes dictionary
    #================================================================================================

    # Topography, Fuel Model, and Vegetation
    slope              : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["slope"]).get(t, y, x)
    aspect             : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["aspect"]).get(t, y, x)
    fuel_model_number  : cy.int   = cy.cast(cy.int, cy.cast(ISpaceTimeCube, space_time_cubes["fuel_model"]).get(t, y, x))
    canopy_cover       : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["canopy_cover"]).get(t, y, x)
    canopy_height      : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["canopy_height"]).get(t, y, x)
    canopy_base_height : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["canopy_base_height"]).get(t, y, x)
    canopy_bulk_density: cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["canopy_bulk_density"]).get(t, y, x)

    # Wind, Surface Moisture, and Foliar Moisture
    wind_speed_10m               : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["wind_speed_10m"]).get(t, y, x)
    upwind_direction             : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["upwind_direction"]).get(t, y, x)
    fuel_moisture_dead_1hr       : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_dead_1hr"]).get(t, y, x)
    fuel_moisture_dead_10hr      : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_dead_10hr"]).get(t, y, x)
    fuel_moisture_dead_100hr     : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_dead_100hr"]).get(t, y, x)
    fuel_moisture_live_herbaceous: cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_live_herbaceous"]).get(t, y, x)
    fuel_moisture_live_woody     : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["fuel_moisture_live_woody"]).get(t, y, x)
    foliar_moisture              : cy.float = cy.cast(ISpaceTimeCube, space_time_cubes["foliar_moisture"]).get(t, y, x)

    # Spread Rate Adjustments (Optional)
    fuel_spread_adjustment   : cy.float = (cy.cast(ISpaceTimeCube, space_time_cubes["fuel_spread_adjustment"]).get(t, y, x)
                                           if "fuel_spread_adjustment" in space_time_cubes
                                           else 1.0)                                         # float >= 0.0
    weather_spread_adjustment: cy.float = (cy.cast(ISpaceTimeCube, space_time_cubes["weather_spread_adjustment"]).get(t, y, x)
                                           if "weather_spread_adjustment" in space_time_cubes
                                           else 1.0)                                         # float >= 0.0
    spread_rate_adjustment   : cy.float = fuel_spread_adjustment * weather_spread_adjustment # float >= 0.0

    #================================================================================================
    # Project a 2D unit vector pointing toward the azimuth onto the slope-tangential plane
    #================================================================================================

    upslope_direction: cy.float = conv.opposite_direction(aspect)
    slope_vector_2d  : vec_xy   = conv.azimuthal_to_cartesian(slope, upslope_direction)
    azimuth_vector_2d: vec_xy   = conv.azimuthal_to_cartesian(1.0, azimuth)
    spread_direction : vec_xyz  = vu.as_unit_vector_3d(vu.to_slope_plane(azimuth_vector_2d, slope_vector_2d))

    #================================================================================================
    # Check whether cell is burnable
    #================================================================================================

    fuel_model      : FuelModel
    maybe_fuel_model: FuelModel|None = fm.fuel_model_table.get(fuel_model_number)

    if maybe_fuel_model:
        fuel_model = maybe_fuel_model

    if (maybe_fuel_model is None or not(fuel_model.burnable)):
        # Encountered unknown or non-burnable fuel model

        #============================================================================================
        # Return zero surface fire behavior in the direction of the azimuth vector
        #============================================================================================

        return {
            "fire_type"         : 0, # unburned
            "spread_rate"       : 0.0,
            "spread_direction"  : spread_direction,
            "fireline_intensity": 0.0,
            "flame_length"      : 0.0,
        }

    else:
        # Encountered burnable fuel model

        #============================================================================================
        # Compute derived parameters
        #============================================================================================

        fuel_moisture               : fclaarr  = (fuel_moisture_dead_1hr,
                                                  fuel_moisture_dead_10hr,
                                                  fuel_moisture_dead_100hr,
                                                  0.0, # fuel_moisture_dead_herbaceous
                                                  fuel_moisture_live_herbaceous,
                                                  fuel_moisture_live_woody) # kg moisture/kg ovendry weight
        fuel_bed_depth              : cy.float = fuel_model.delta                      # ft
        heat_of_combustion          : cy.float = conv.Btu_lb_to_kJ_kg(fuel_model.h[0]) # kJ/kg
        estimated_fine_fuel_moisture: cy.float = fuel_moisture_dead_1hr                # kg moisture/kg ovendry weight

        #============================================================================================
        # Calculate midflame wind speed
        #============================================================================================

        # Convert from 10m wind speed to 20ft wind speed
        wind_speed_20ft: cy.float = conv.wind_speed_10m_to_wind_speed_20ft(wind_speed_10m) # km/hr

        # Convert 20ft wind speed from km/hr to m/min
        wind_speed_20ft_m_min: cy.float = conv.km_hr_to_m_min(wind_speed_20ft) # m/min

        # Convert from 20ft wind speed to midflame wind speed in m/min
        midflame_wind_speed: cy.float = sf.calc_midflame_wind_speed(wind_speed_20ft_m_min,       # m/min
                                                                    fuel_bed_depth,              # ft
                                                                    conv.m_to_ft(canopy_height), # ft
                                                                    canopy_cover)                # 0-1

        #============================================================================================
        # Calculate surface fire behavior in the direction of maximum spread
        #============================================================================================

        # Apply fuel moisture to fuel model
        moisturized_fuel_model: FuelModel = fm.moisturize(fuel_model, fuel_moisture)

        # TODO: Memoize calc_surface_fire_behavior_no_wind_no_slope
        # Calculate no-wind-no-slope surface fire behavior
        surface_fire_min: FireBehaviorMin = sf.calc_surface_fire_behavior_no_wind_no_slope(moisturized_fuel_model,
                                                                                           spread_rate_adjustment)

        # Calculate surface fire behavior in the direction of maximum spread
        surface_fire_max: FireBehaviorMax = sf.calc_surface_fire_behavior_max(surface_fire_min,
                                                                              midflame_wind_speed,
                                                                              upwind_direction,
                                                                              slope,
                                                                              aspect,
                                                                              use_wind_limit,
                                                                              surface_lw_ratio_model)

        #============================================================================================
        # Calculate surface fire behavior in the direction of the azimuth vector
        #============================================================================================

        surface_fire_azimuth: SpreadBehavior = sf.calc_surface_fire_behavior_in_direction(surface_fire_max,
                                                                                          spread_direction)

        #============================================================================================
        # Determine whether the surface fire transitions to a crown fire
        #============================================================================================

        if cf.van_wagner_crown_fire_initiation(surface_fire_azimuth.fireline_intensity,
                                               canopy_cover,
                                               canopy_base_height,
                                               foliar_moisture):

            #========================================================================================
            # Calculate crown fire behavior in the direction of maximum spread
            #========================================================================================

            crown_fire_max: FireBehaviorMax = cf.calc_crown_fire_behavior_max(canopy_height, canopy_base_height,
                                                                              canopy_bulk_density, heat_of_combustion,
                                                                              estimated_fine_fuel_moisture,
                                                                              wind_speed_10m, upwind_direction,
                                                                              slope, aspect, crown_max_lw_ratio)

            #========================================================================================
            # Calculate crown fire behavior in the direction of the azimuth vector
            #========================================================================================

            crown_fire_azimuth: SpreadBehavior = cf.calc_crown_fire_behavior_in_direction(crown_fire_max,
                                                                                          spread_direction)

            #========================================================================================
            # Calculate combined fire behavior in the direction of the azimuth vector
            #========================================================================================

            combined_fire_azimuth: SpreadBehavior = cf.calc_combined_fire_behavior(surface_fire_azimuth,
                                                                                   crown_fire_azimuth)

            #========================================================================================
            # Return the combined fire behavior in the direction of the azimuth vector
            #========================================================================================

            return SpreadBehavior_to_dict(combined_fire_azimuth)

        else:

            #========================================================================================
            # Return the surface fire behavior in the direction of the azimuth vector
            #========================================================================================

            return SpreadBehavior_to_dict(surface_fire_azimuth)


# TODO: Make a more efficient version that avoids space_time_cubes dictionary lookups for each cell
@cy.ccall
def burn_all_cells_toward_azimuth(space_time_cubes      : dict[str, ISpaceTimeCube],
                                  azimuth               : cy.float,
                                  t                     : pyidx,
                                  y_range               : tuple[pyidx, pyidx]|None = None,
                                  x_range               : tuple[pyidx, pyidx]|None = None,
                                  use_wind_limit        : cy.bint = True,
                                  surface_lw_ratio_model: str = "behave",
                                  crown_max_lw_ratio    : cy.float = 1e10) -> dict:
    """
    Given these inputs:
    - space_time_cubes             :: dictionary of (Lazy)SpaceTimeCube objects with these cell types
      - slope                         :: rise/run
      - aspect                        :: degrees clockwise from North
      - fuel_model                    :: integer index in fm.fuel_model_table
      - canopy_cover                  :: 0-1
      - canopy_height                 :: m
      - canopy_base_height            :: m
      - canopy_bulk_density           :: kg/m^3
      - wind_speed_10m                :: km/hr
      - upwind_direction              :: degrees clockwise from North
      - fuel_moisture_dead_1hr        :: kg moisture/kg ovendry weight
      - fuel_moisture_dead_10hr       :: kg moisture/kg ovendry weight
      - fuel_moisture_dead_100hr      :: kg moisture/kg ovendry weight
      - fuel_moisture_live_herbaceous :: kg moisture/kg ovendry weight
      - fuel_moisture_live_woody      :: kg moisture/kg ovendry weight
      - foliar_moisture               :: kg moisture/kg ovendry weight
      - fuel_spread_adjustment        :: float >= 0.0 (Optional: defaults to 1.0)
      - weather_spread_adjustment     :: float >= 0.0 (Optional: defaults to 1.0)
    - azimuth                      :: degrees clockwise from North on the horizontal plane
    - t                            :: temporal integer index into the ISpaceTimeCube objects
    - y_range                      :: (min_y, max_y) spatial integer index into the ISpaceTimeCube objects (Optional)
    - x_range                      :: (min_x, max_x) spatial integer index into the ISpaceTimeCube objects (Optional)
    - use_wind_limit               :: boolean (Optional)
    - surface_lw_ratio_model       :: "rothermel" or "behave" (Optional)
    - crown_max_lw_ratio           :: float > 0.0 (Optional)

    return a dictionary with these keys:
    - fire_type          :: 2D byte array (0=unburned, 1=surface, 2=passive_crown, 3=active_crown)
    - spread_rate        :: 2D float array (m/min)
    - spread_direction   :: 2D float array (degrees clockwise from North)
    - fireline_intensity :: 2D float array (kW/m)
    - flame_length       :: 2D float array (m)
    """
    slope_cube  : ISpaceTimeCube      = cy.cast(ISpaceTimeCube, space_time_cubes["slope"])
    bands       : pyidx               = slope_cube.shape[0]
    rows        : pyidx               = slope_cube.shape[1]
    cols        : pyidx               = slope_cube.shape[2]
    grid_shape  : tuple[int, int]     = (rows, cols)
    y_range_real: tuple[pyidx, pyidx] = to_positive_index_range(y_range, rows)
    x_range_real: tuple[pyidx, pyidx] = to_positive_index_range(x_range, cols)
    min_y       : pyidx               = y_range_real[0]
    max_y       : pyidx               = y_range_real[1]
    min_x       : pyidx               = x_range_real[0]
    max_x       : pyidx               = x_range_real[1]

    if not(0 <= t < bands):
        raise ValueError("The t value is out of range of the space_time_cubes.")

    if not(0 <= min_y < max_y <= rows):
        raise ValueError("The y_range values are out of range of the space_time_cubes.")

    if not(0 <= min_x < max_x <= rows):
        raise ValueError("The x_range values are out of range of the space_time_cubes.")

    fire_type_matrix          : ndarray         = np.zeros(grid_shape, dtype="uint8")
    spread_rate_matrix        : ndarray         = np.zeros(grid_shape, dtype="float32")
    spread_direction_matrix   : ndarray         = np.zeros(grid_shape, dtype="float32")
    fireline_intensity_matrix : ndarray         = np.zeros(grid_shape, dtype="float32")
    flame_length_matrix       : ndarray         = np.zeros(grid_shape, dtype="float32")
    fire_type_memview         : cy.uchar[:,::1] = fire_type_matrix
    spread_rate_memview       : cy.float[:,::1] = spread_rate_matrix
    spread_direction_memview  : cy.float[:,::1] = spread_direction_matrix
    fireline_intensity_memview: cy.float[:,::1] = fireline_intensity_matrix
    flame_length_memview      : cy.float[:,::1] = flame_length_matrix

    y                    : pyidx
    x                    : pyidx
    space_time_coordinate: coord_tyx
    for y in range(min_y, max_y):
        for x in range(min_x, max_x):
            space_time_coordinate          = (t, y, x)
            spread_behavior                = burn_cell_toward_azimuth(space_time_cubes,
                                                                      space_time_coordinate,
                                                                      azimuth,
                                                                      use_wind_limit,
                                                                      surface_lw_ratio_model,
                                                                      crown_max_lw_ratio)
            fire_type_memview[y,x]          = spread_behavior["fire_type"]
            spread_rate_memview[y,x]        = spread_behavior["spread_rate"]
            spread_direction_memview[y,x]   = vu.spread_direction_vector_to_angle(spread_behavior["spread_direction"])
            fireline_intensity_memview[y,x] = spread_behavior["fireline_intensity"]
            flame_length_memview[y,x]       = spread_behavior["flame_length"]

    return {
        "fire_type"         : fire_type_matrix,
        "spread_rate"       : spread_rate_matrix,
        "spread_direction"  : spread_direction_matrix,
        "fireline_intensity": fireline_intensity_matrix,
        "flame_length"      : flame_length_matrix,
    }
# burn-cells-toward-azimuth ends here



# [[file:../../org/pyretechnics.org::py-types][py-types]]
#==============================================================
# Runtime-defined type aliases
#==============================================================

import cython as cy

pyidx            = cy.typedef(cy.Py_ssize_t)
vec_xy           = cy.typedef(tuple) # tuple[cy.float, cy.float]
vec_xyz          = cy.typedef(tuple) # tuple[cy.float, cy.float, cy.float]
coord_yx         = cy.typedef(tuple) # tuple[pyidx, pyidx]
coord_tyx        = cy.typedef(tuple) # tuple[pyidx, pyidx, pyidx]
fcatarr          = cy.typedef(tuple) # tuple[cy.float, cy.float]
fclaarr          = cy.typedef(tuple) # tuple[cy.float, cy.float, cy.float, cy.float, cy.float, cy.float]
CompactFuelModel = cy.typedef(tuple) # tuple[cy.float, cy.float, cy.float, cy.float, cy.float, cy.float,
                                     #       cy.float, cy.float, cy.float, cy.float, cy.float, cy.float, cy.float]

FuelModel = cy.struct(
    number               = cy.int,
    delta                = cy.float,
    M_x                  = fclaarr,
    M_f                  = fclaarr,
    w_o                  = fclaarr,
    sigma                = fclaarr,
    h                    = fclaarr,
    rho_p                = fclaarr,
    S_T                  = fclaarr,
    S_e                  = fclaarr,
    dynamic              = cy.bint,
    burnable             = cy.bint,
    exp_A_sigma          = fclaarr,
    firemod_size_classes = fclaarr,
    f_ij                 = fclaarr,
    f_i                  = fcatarr,
    g_ij                 = fclaarr,
)

ProjectedVectors = cy.struct(
    wind_vector_3d  = vec_xyz,
    slope_vector_3d = vec_xyz,
)

FireBehaviorMin = cy.struct(
    base_spread_rate         = cy.float,
    base_fireline_intensity  = cy.float,
    max_effective_wind_speed = cy.float,
    _phiS_G                  = cy.float,
    _phiW_scalr              = cy.float,
    _phiW_expnt              = cy.float,
    _ws_scalr                = cy.float,
    _ws_expnt                = cy.float,
)

FireBehaviorMax = cy.struct(
    max_fire_type          = cy.int,
    max_spread_rate        = cy.float,
    max_spread_direction   = vec_xyz,
    max_fireline_intensity = cy.float,
    max_flame_length       = cy.float,
    length_to_width_ratio  = cy.float,
    eccentricity           = cy.float,
    critical_spread_rate   = cy.float,
)

SpreadBehavior = cy.struct(
    dphi_dt            = cy.float,
    fire_type          = cy.int,
    spread_rate        = cy.float,
    spread_direction   = vec_xyz,
    fireline_intensity = cy.float,
    flame_length       = cy.float,
)

CrownSpreadInfo = cy.struct(
    fire_type            = cy.int,
    spread_rate          = cy.float,
    critical_spread_rate = cy.float,
)

SpotConfig = cy.struct(
    random_seed                  = cy.long,
    firebrands_per_unit_heat     = cy.float,
    downwind_distance_mean       = cy.float,
    fireline_intensity_exponent  = cy.float,
    wind_speed_exponent          = cy.float,
    downwind_variance_mean_ratio = cy.float,
    crosswind_distance_stdev     = cy.float,
    decay_distance               = cy.float,
)

JumpDistribution = cy.struct(
    # Downwind LogNormal params
    # Formally, we have ln(downwind_jump / 1m) ~ Normal(mu = mu_x, sigma = sigma_x)
    mu_x    = cy.float, # dimensionless (log-space)
    sigma_x = cy.float, # dimensionless (log-space)
    # Crosswind normal params
    # Formally, we have crosswind_jump ~ Normal(mu = 0, sigma = sigma_y)
    sigma_y = cy.float, # meters
)

# Pre-computed coefficients to apply elliptical wavelet math as fast as possible
# once the phi gradient information is available.
# See `pyretechnics.eulerian_level_set.dphi_dt_from_partialed_wavelet`.
PartialedEllWavelet = cy.struct(
    Vh_3d = vec_xyz,  # Heading spread rate vector (m/min)
    ewc_A = cy.float, # Dimensionless elliptical wavelet coefficient (<= 0)
    ewc_B = cy.float, # Dimensionless elliptical wavelet coefficient (<= 0)
    ewc_C = cy.float, # Dimensionless elliptical wavelet coefficient (>= 0)
)

CellInputs = cy.struct(
    slope                         = cy.float,
    aspect                        = cy.float,
    fuel_model_number             = cy.float,
    canopy_cover                  = cy.float,
    canopy_height                 = cy.float,
    canopy_base_height            = cy.float,
    canopy_bulk_density           = cy.float,
    wind_speed_10m                = cy.float,
    upwind_direction              = cy.float,
    fuel_moisture_dead_1hr        = cy.float,
    fuel_moisture_dead_10hr       = cy.float,
    fuel_moisture_dead_100hr      = cy.float,
    fuel_moisture_live_herbaceous = cy.float,
    fuel_moisture_live_woody      = cy.float,
    foliar_moisture               = cy.float,
    fuel_spread_adjustment        = cy.float,
    weather_spread_adjustment     = cy.float,
)

# Pre-computed information required to compute dphi/dt once the phi
# gradient is known. Derived from the surface and crown wavelets.
#
# NOTE: The reason to make this a small struct stored in an array is
#       efficiency - we want the CPU to have a low cache miss rate.
#
# NOTE: A significant benefit of this architecture is that it's
#       Rothermel-agnostic. EllipticalInfo could conceivably be
#       implemented using variants of the Rothermel model. This can be
#       valuable to give flexibility to users.
EllipticalInfo = cy.struct(
    cell_index           = coord_yx,
    elevation_gradient   = vec_xy,
    surface_wavelet      = PartialedEllWavelet,
    crown_wavelet        = PartialedEllWavelet,
    crowning_spread_rate = cy.float, # Surface spread rate at which crowning occurs
)

# Some data saved during the 1st Runge-Kutta pass.
Pass1CellOutput = cy.struct(
    cell_index      = coord_yx,
    phi_gradient_xy = vec_xy,
    dphi_dt_flim    = cy.float, # Flux-limited dphi/dt (phi/min, <= 0).
    phi_old         = cy.float,
)
# py-types ends here
